@@ -2,7 +2,7 @@
  * dpkg-deb - construction and deconstruction of *.deb archives
  * extract.c - extracting archives
  *
- * Copyright (C) 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
@@ -19,6 +19,9 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 #include <config.h>
+#include <compat.h>
+
+#include <dpkg/i18n.h>
 
 #include <stdio.h>
 #include <string.h>
@@ -38,9 +41,12 @@
 #include <zlib.h>
 #endif
 
-#include <dpkg.h>
-#include <dpkg-deb.h>
-#include <myopt.h>
+#include <dpkg/dpkg.h>
+#include <dpkg/buffer.h>
+#include <dpkg/subproc.h>
+#include <dpkg/myopt.h>
+
+#include "dpkg-deb.h"
 
 static void movecontrolfiles(const char *thing) {
   char buf[200];
@@ -48,24 +54,27 @@ static void movecontrolfiles(const char *thing) {
   
   sprintf(buf, "mv %s/* . && rmdir %s", thing, thing);
   if (!(c1= m_fork())) {
-    execlp("sh","sh","-c",buf,(char*)0); ohshite(_("failed to exec sh -c mv foo/* &c"));
+    execlp("sh", "sh", "-c", buf, NULL);
+    ohshite(_("failed to exec sh -c mv foo/* &c"));
   }
   waitsubproc(c1,"sh -c mv foo/* &c",0);
 }
 
-static void readfail(FILE *a, const char *filename, const char *what) NONRETURNING;
+static void readfail(FILE *a, const char *filename, const char *what) DPKG_ATTR_NORET;
 static void readfail(FILE *a, const char *filename, const char *what) {
   if (ferror(a)) {
-    ohshite(_("error reading %s from %.255s"),what,filename);
+    ohshite(_("error reading %s from file %.255s"), what, filename);
   } else {
     ohshit(_("unexpected end of file in %s in %.255s"),what,filename);
   }
 }
 
-static unsigned long parseheaderlength(const char *inh, size_t len,
-                                       const char *fn, const char *what) {
+static size_t
+parseheaderlength(const char *inh, size_t len,
+                  const char *fn, const char *what)
+{
   char lintbuf[15];
-  unsigned long r;
+  ssize_t r;
   char *endp;
 
   if (memchr(inh,0,len))
@@ -73,11 +82,31 @@ static unsigned long parseheaderlength(const char *inh, size_t len,
   assert(sizeof(lintbuf) > len);
   memcpy(lintbuf,inh,len);
   lintbuf[len]= ' ';
-  *strchr(lintbuf,' ')= 0;
-  r= strtoul(lintbuf,&endp,10);
+  *strchr(lintbuf, ' ') = '\0';
+  r = strtol(lintbuf, &endp, 10);
+  if (r < 0)
+    ohshit(_("file `%.250s' is corrupt - negative member length %zi"), fn, r);
   if (*endp)
     ohshit(_("file `%.250s' is corrupt - bad digit (code %d) in %s"),fn,*endp,what);
-  return r;
+  return (size_t)r;
+}
+
+static void
+safe_fflush(FILE *f)
+{
+#if defined(__GLIBC__) && (__GLIBC__ == 2) && (__GLIBC_MINOR__ > 0)
+  /* XXX: Glibc 2.1 and some versions of Linux want to make fflush()
+   * move the current fpos. Remove this code some time. */
+  fpos_t fpos;
+
+  if (fgetpos(f, &fpos))
+    ohshit(_("failed getting the current file position"));
+  fflush(f);
+  if (fsetpos(f, &fpos))
+    ohshit(_("failed setting the current file position"));
+#else
+  fflush(f);
+#endif
 }
 
 void extracthalf(const char *debar, const char *directory,
@@ -88,7 +117,7 @@ void extracthalf(const char *debar, const char *directory,
   size_t ctrllennum, memberlen= 0;
   int dummy, l= 0;
   pid_t c1=0,c2,c3;
-  unsigned char *ctrlarea= 0;
+  unsigned char *ctrlarea = NULL;
   int p1[2], p2[2];
   FILE *ar, *pi;
   struct stat stab;
@@ -96,10 +125,7 @@ void extracthalf(const char *debar, const char *directory,
   char *cur;
   struct ar_hdr arh;
   int readfromfd, oldformat= 0, header_done, adminmember;
-#if defined(__GLIBC__) && (__GLIBC__ == 2) && (__GLIBC_MINOR__ > 0)
-  fpos_t fpos;
-#endif
-  enum compression_type compress_type = GZ;
+  enum compress_type compress_type = compress_type_gzip;
   
   ar= fopen(debar,"r"); if (!ar) ohshite(_("failed to read archive `%.255s'"),debar);
   if (fstat(fileno(ar),&stab)) ohshite(_("failed to fstat archive"));
@@ -116,9 +142,7 @@ void extracthalf(const char *debar, const char *directory,
       if (memcmp(arh.ar_fmag,ARFMAG,sizeof(arh.ar_fmag)))
         ohshit(_("file `%.250s' is corrupt - bad magic at end of first header"),debar);
       memberlen= parseheaderlength(arh.ar_size,sizeof(arh.ar_size),
-                                   debar,"member length");
-      if (memberlen<0)
-        ohshit(_("file `%.250s' is corrupt - negative member length %zi"),debar,memberlen);
+                                   debar, _("member length"));
       if (!header_done) {
         if (memcmp(arh.ar_name,"debian-binary   ",sizeof(arh.ar_name)) &&
 	    memcmp(arh.ar_name,"debian-binary/   ",sizeof(arh.ar_name)))
@@ -126,18 +150,18 @@ void extracthalf(const char *debar, const char *directory,
         infobuf= m_malloc(memberlen+1);
         if (fread(infobuf,1, memberlen + (memberlen&1), ar) != memberlen + (memberlen&1))
           readfail(ar,debar,_("header info member"));
-        infobuf[memberlen]= 0;
+        infobuf[memberlen] = '\0';
         cur= strchr(infobuf,'\n');
         if (!cur) ohshit(_("archive has no newlines in header"));
-        *cur= 0;
+        *cur = '\0';
         cur= strchr(infobuf,'.');
         if (!cur) ohshit(_("archive has no dot in version number"));
-        *cur= 0;
+        *cur = '\0';
         if (strcmp(infobuf,"2"))
           ohshit(_("archive version %.250s not understood, get newer dpkg-deb"), infobuf);
         *cur= '.';
         strncpy(versionbuf,infobuf,sizeof(versionbuf));
-        versionbuf[sizeof(versionbuf)-1]= 0;
+        versionbuf[sizeof(versionbuf) - 1] = '\0';
         header_done= 1;
       } else if (arh.ar_name[0] == '_') {
           /* Members with `_' are noncritical, and if we don't understand them
@@ -152,15 +176,19 @@ void extracthalf(const char *debar, const char *directory,
 	  if (!memcmp(arh.ar_name,DATAMEMBER_GZ,sizeof(arh.ar_name)) ||
 	      !memcmp(arh.ar_name,DATAMEMBER_COMPAT_GZ,sizeof(arh.ar_name))) {
 	    adminmember= 0;
-	    compress_type= GZ;
+	    compress_type = compress_type_gzip;
 	  } else if (!memcmp(arh.ar_name,DATAMEMBER_BZ2,sizeof(arh.ar_name)) ||
 		     !memcmp(arh.ar_name,DATAMEMBER_COMPAT_BZ2,sizeof(arh.ar_name))) {
 	    adminmember= 0;
-	    compress_type= BZ2;
+	    compress_type = compress_type_bzip2;
+	  } else if (!memcmp(arh.ar_name, DATAMEMBER_LZMA, sizeof(arh.ar_name)) ||
+		     !memcmp(arh.ar_name, DATAMEMBER_COMPAT_LZMA, sizeof(arh.ar_name))) {
+	    adminmember = 0;
+	    compress_type = compress_type_lzma;
 	  } else if (!memcmp(arh.ar_name,DATAMEMBER_CAT,sizeof(arh.ar_name)) ||
 		     !memcmp(arh.ar_name,DATAMEMBER_COMPAT_CAT,sizeof(arh.ar_name))) {
 	    adminmember= 0;
-	    compress_type= CAT;
+	    compress_type = compress_type_cat;
 	  } else {
             ohshit(_("file `%.250s' contains ununderstood data member %.*s, giving up"),
                    debar, (int)sizeof(arh.ar_name), arh.ar_name);
@@ -179,61 +207,60 @@ void extracthalf(const char *debar, const char *directory,
       }
     }
 
-    if (admininfo >= 2)
-      if (printf(_(" new debian package, version %s.\n"
-                 " size %ld bytes: control archive= %zi bytes.\n"),
-                 versionbuf, (long)stab.st_size, ctrllennum) == EOF ||
-          fflush(stdout)) werr("stdout");
-    
+    if (admininfo >= 2) {
+      printf(_(" new debian package, version %s.\n"
+               " size %ld bytes: control archive= %zi bytes.\n"),
+             versionbuf, (long)stab.st_size, ctrllennum);
+      m_output(stdout, _("<standard output>"));
+    }
   } else if (!strncmp(versionbuf,"0.93",4) &&
              sscanf(versionbuf,"%f%c%d",&versionnum,&nlc,&dummy) == 2 &&
              nlc == '\n') {
     
     oldformat= 1;
-    l= strlen(versionbuf); if (l && versionbuf[l-1]=='\n') versionbuf[l-1]=0;
+    l = strlen(versionbuf);
+    if (l && versionbuf[l - 1] == '\n')
+      versionbuf[l - 1] = '\0';
     if (!fgets(ctrllenbuf,sizeof(ctrllenbuf),ar))
-      readfail(ar,debar,_("ctrl information length"));
+      readfail(ar, debar, _("control information length"));
     if (sscanf(ctrllenbuf,"%zi%c%d",&ctrllennum,&nlc,&dummy) !=2 || nlc != '\n')
-      ohshit(_("archive has malformatted ctrl len `%s'"),ctrllenbuf);
+      ohshit(_("archive has malformatted control length `%s'"), ctrllenbuf);
 
-    if (admininfo >= 2)
-      if (printf(_(" old debian package, version %s.\n"
-                 " size %ld bytes: control archive= %zi, main archive= %ld.\n"),
-                 versionbuf, (long)stab.st_size, ctrllennum,
-                 (long) (stab.st_size - ctrllennum - strlen(ctrllenbuf) - l)) == EOF ||
-          fflush(stdout)) werr("stdout");
-    
-    ctrlarea= malloc(ctrllennum); if (!ctrlarea) ohshite("malloc ctrlarea failed");
+    if (admininfo >= 2) {
+      printf(_(" old debian package, version %s.\n"
+               " size %ld bytes: control archive= %zi, main archive= %ld.\n"),
+             versionbuf, (long)stab.st_size, ctrllennum,
+             (long) (stab.st_size - ctrllennum - strlen(ctrllenbuf) - l));
+      m_output(stdout, _("<standard output>"));
+    }
+
+    ctrlarea = m_malloc(ctrllennum);
 
     errno=0; if (fread(ctrlarea,1,ctrllennum,ar) != ctrllennum)
-      readfail(ar,debar,_("ctrlarea"));
+      readfail(ar, debar, _("control area"));
 
   } else {
     
     if (!strncmp(versionbuf,"!<arch>",7)) {
-      if (fprintf(stderr,
-                  _("dpkg-deb: file looks like it might be an archive which has been\n"
-                  "dpkg-deb:    corrupted by being downloaded in ASCII mode\n"))
-          == EOF) werr("stderr");
+      fprintf(stderr,
+              _("dpkg-deb: file looks like it might be an archive which has been\n"
+                "dpkg-deb:    corrupted by being downloaded in ASCII mode\n"));
     }
 
     ohshit(_("`%.255s' is not a debian format archive"),debar);
 
   }
 
-#if defined(__GLIBC__) && (__GLIBC__ == 2) && (__GLIBC_MINOR__ > 0)
-  if(fgetpos(ar, &fpos)) ohshit(_("fgetpos failed"));
-#endif
-  fflush(ar);
-#if defined(__GLIBC__) && (__GLIBC__ == 2) && (__GLIBC_MINOR__ > 0)
-  if(fsetpos(ar, &fpos)) ohshit(_("fsetpos failed"));
-#endif
+  safe_fflush(ar);
+
   if (oldformat) {
     if (admininfo) {
       m_pipe(p1);
       if (!(c1= m_fork())) {
         close(p1[0]);
-        if (!(pi= fdopen(p1[1],"w"))) ohshite(_("failed to fdopen p1 in paste"));
+	pi = fdopen(p1[1], "w");
+	if (!pi)
+	  ohshite(_("failed to open pipe descriptor `1' in paste"));
         errno=0; if (fwrite(ctrlarea,1,ctrllennum,pi) != ctrllennum)
           ohshit(_("failed to write to gzip -dc"));
         if (fclose(pi)) ohshit(_("failed to close gzip -dc"));
@@ -284,20 +311,25 @@ void extracthalf(const char *debar, const char *directory,
   if (taroption) {
     if (!(c3= m_fork())) {
       char buffer[30+2];
-      if(strlen(taroption) > 30) internerr(taroption);
+      if (strlen(taroption) > 30)
+        internerr("taroption is too long '%s'", taroption);
       strcpy(buffer, taroption);
       strcat(buffer, "f");
       m_dup2(p2[0],0);
       close(p2[0]);
-      execlp(TAR,"tar",buffer,"-",(char*)0);
+
+      unsetenv("TAR_OPTIONS");
+
+      execlp(TAR, "tar", buffer, "-", NULL);
       ohshite(_("failed to exec tar"));
     }
     close(p2[0]);
     waitsubproc(c3,"tar",0);
   }
   
-  waitsubproc(c2,"<decompress>",PROCPIPE);
-  if (c1 != -1) waitsubproc(c1,"paste",0);
+  waitsubproc(c2, _("<decompress>"), PROCPIPE);
+  if (c1 != -1)
+    waitsubproc(c1, _("paste"), 0);
   if (oldformat && admininfo) {
     if (versionnum == 0.931F) {
       movecontrolfiles(OLDOLDDEBDIR);
@@ -331,7 +363,7 @@ void do_fsystarfile(const char *const *argv) {
     badusage(_("--%s needs a .deb filename argument"),cipaction->olong);
   if (*argv)
     badusage(_("--%s takes only one argument (.deb filename)"),cipaction->olong);
-  extracthalf(debar,0,0,0);
+  extracthalf(debar, NULL, NULL, 0);
 }
    
 void do_control(const char *const *argv) { controlextractvextract(1, "x", argv); }
