@@ -1,4 +1,4 @@
-# Copyright 2007 Raphael Hertzog <hertzog@debian.org>
+# Copyright © 2007-2009 Raphael Hertzog <hertzog@debian.org>
 #
 # This program is free software; you may redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,11 +18,11 @@
 #########################################################################
 # Several parts are inspired by lib/Dep.pm from lintian (same license)
 #
-# Copyright (C) 1998 Richard Braakman
-# Portions Copyright (C) 1999 Darren Benham
-# Portions Copyright (C) 2000 Sean 'Shaleh' Perry
-# Portions Copyright (C) 2004 Frank Lichtenheld
-# Portions Copyright (C) 2006 Russ Allbery
+# Copyright © 1998 Richard Braakman
+# Portions Copyright © 1999 Darren Benham
+# Portions Copyright © 2000 Sean 'Shaleh' Perry
+# Portions Copyright © 2004 Frank Lichtenheld
+# Portions Copyright © 2006 Russ Allbery
 
 package Dpkg::Deps;
 
@@ -70,10 +70,10 @@ use warnings;
 
 use Dpkg::Version qw(compare_versions);
 use Dpkg::Arch qw(get_host_arch);
-use Dpkg::ErrorHandling qw(warning);
+use Dpkg::ErrorHandling;
 use Dpkg::Gettext;
 
-our @ISA = qw(Exporter);
+use base qw(Exporter);
 our @EXPORT_OK = qw(@pkg_dep_fields @src_dep_fields %dep_field_type
 		    %relation_ordering);
 
@@ -116,6 +116,7 @@ The arguments can also be undef in case there's no explicit architecture
 restriction.
 
 =cut
+
 sub arch_is_superset {
     my ($p, $q) = @_;
     my $p_arch_neg = defined($p) && $p->[0] =~ /^!/;
@@ -186,6 +187,7 @@ dependency. It returns 0 if the "p" dependency implies that "q" is
 not satisfied. It returns undef when there's no implication.
 
 =cut
+
 sub version_implies {
     my ($rel_p, $v_p, $rel_q, $v_q) = @_;
 
@@ -296,6 +298,7 @@ this when parsing non-dependency fields like Conflicts (see
 =back
 
 =cut
+
 sub parse {
     my $dep_line = shift;
     my %options = (@_);
@@ -304,13 +307,17 @@ sub parse {
     $options{host_arch} = get_host_arch() if not exists $options{host_arch};
     $options{union} = 0 if not exists $options{union};
 
+    # Strip trailing/leading spaces
+    $dep_line =~ s/^\s+//;
+    $dep_line =~ s/\s+$//;
+
     my @dep_list;
     foreach my $dep_and (split(/\s*,\s*/m, $dep_line)) {
         my @or_list = ();
         foreach my $dep_or (split(/\s*\|\s*/m, $dep_and)) {
 	    my $dep_simple = Dpkg::Deps::Simple->new($dep_or);
 	    if (not defined $dep_simple->{package}) {
-		warning(sprintf(_g("can't parse dependency %s"), $dep_and));
+		warning(_g("can't parse dependency %s"), $dep_or);
 		return undef;
 	    }
 	    $dep_simple->{arches} = undef if not $options{use_arch};
@@ -335,7 +342,13 @@ sub parse {
     } else {
 	$dep_and = Dpkg::Deps::AND->new();
     }
-    $dep_and->add($_) foreach (@dep_list);
+    foreach my $dep (@dep_list) {
+        if ($options{union} and not $dep->isa("Dpkg::Deps::Simple")) {
+            warning(_g("an union dependency can only contain simple dependencies"));
+            return undef;
+        }
+        $dep_and->add($dep);
+    }
     return $dep_and;
 }
 
@@ -347,6 +360,7 @@ dumping.
 =back
 
 =cut
+
 sub compare {
     my ($a, $b) = @_;
     return -1 if $a->is_empty();
@@ -507,7 +521,7 @@ use warnings;
 
 use Dpkg::Arch qw(debarch_is);
 use Dpkg::Version qw(compare_versions);
-use Dpkg::ErrorHandling qw(internerr);
+use Dpkg::ErrorHandling;
 use Dpkg::Gettext;
 
 sub new {
@@ -541,7 +555,7 @@ sub parse {
                 \s* \]                      # closing bracket
               )?                            # end of optional architecture
 	      \s*$			    # trailing spaces at end
-            /mx;
+            /x;
     $self->{package} = $1;
     $self->{relation} = $2;
     $self->{version} = $3;
@@ -623,7 +637,8 @@ sub implies {
 	}
 	return $res;
     } else {
-	internerr(sprintf(_g("Dpkg::Deps::Simple can't evaluate implication with a %s!"), ref($o)));
+	internerr("Dpkg::Deps::Simple can't evaluate implication with a %s!",
+	          ref($o));
     }
 }
 
@@ -774,10 +789,11 @@ Add a new dependency object at the end of the list.
 =back
 
 =cut
+
 use strict;
 use warnings;
 
-use Dpkg::ErrorHandling qw(internerr);
+use Dpkg::ErrorHandling;
 
 sub new {
     my $this = shift;
@@ -912,9 +928,21 @@ WHILELOOP:
 	my $dep = shift @{$self->{list}};
 	my $eval = $dep->get_evaluation($facts);
 	next if defined($eval) and $eval == 1;
-	foreach my $odep (@knowndeps, @new, @{$self->{list}}) {
+	foreach my $odep (@knowndeps, @new) {
 	    next WHILELOOP if $odep->implies($dep);
 	}
+        # When a dependency is implied by another dependency that
+        # follows, then invert them
+        # "a | b, c, a"  becomes "a, c" and not "c, a"
+        my $i = 0;
+	foreach my $odep (@{$self->{list}}) {
+            if (defined $odep and $odep->implies($dep)) {
+                splice @{$self->{list}}, $i, 1;
+                unshift @{$self->{list}}, $odep;
+                next WHILELOOP;
+            }
+            $i++;
+	}
 	push @new, $dep;
     }
     $self->{list} = [ @new ];
@@ -1090,6 +1118,7 @@ packages provided (by the set of installed packages).
 Create a new object.
 
 =cut
+
 use strict;
 use warnings;
 
@@ -1108,6 +1137,7 @@ undefined we know that the package is installed but we don't know which
 version it is.
 
 =cut
+
 sub add_installed_package {
     my ($self, $pkg, $ver) = @_;
     $self->{pkg}{$pkg} = $ver;
@@ -1120,6 +1150,7 @@ and $version correspond to the associated relation given in the Provides
 field. This might be used in the future for versioned provides.
 
 =cut
+
 sub add_provided_package {
     my ($self, $pkg, $rel, $ver, $by) = @_;
     if (not exists $self->{virtualpkg}{$pkg}) {
@@ -1138,6 +1169,7 @@ listed as [ $provider, $relation, $version ]).
 =back
 
 =cut
+
 sub check_package {
     my ($self, $pkg) = @_;
     if (exists $self->{pkg}{$pkg}) {
