@@ -3,11 +3,12 @@ package Dpkg::Fields;
 use strict;
 use warnings;
 
-use Exporter;
+use base qw(Exporter);
+use Dpkg::Gettext;
+use Dpkg::ErrorHandling;
 use Dpkg::Deps qw(@src_dep_fields @pkg_dep_fields);
 
-our @ISA = qw(Exporter);
-our @EXPORT_OK = qw(capit %control_src_fields %control_pkg_fields
+our @EXPORT_OK = qw(capit unknown %control_src_fields %control_pkg_fields
     $control_src_field_regex $control_pkg_field_regex);
 our %EXPORT_TAGS = ('list' => [qw(%control_src_fields %control_pkg_fields
 			$control_src_field_regex $control_pkg_field_regex)]);
@@ -22,7 +23,7 @@ $control_src_fields{$_} = 1 foreach (qw(Bugs Dm-Upload-Allowed
 $control_src_fields{$_} = 1 foreach (@src_dep_fields);
 $control_pkg_fields{$_} = 1 foreach (qw(Architecture Bugs Description Essential
     Homepage Installer-Menu-Item Kernel-Version Package Package-Type
-    Priority Section Subarchitecture Tag));
+    Priority Section Subarchitecture Tag Multi-Arch));
 $control_pkg_fields{$_} = 1 foreach (@pkg_dep_fields);
 
 our $control_src_field_regex = "(?:" . join("|", keys %control_src_fields) . ")";
@@ -34,6 +35,14 @@ sub capit {
     return join '-', @pieces;
 }
 
+sub unknown($$)
+{
+    my ($field, $desc) = @_;
+
+    warning(_g("unknown information field '%s' in input data in %s"),
+            $field, $desc);
+}
+
 package Dpkg::Fields::Object;
 
 =head1 OTHER OBJECTS
@@ -46,10 +55,11 @@ Dpkg::Fields::capit). It also stores the order in which fields have been
 added in order to be able to dump them in the same order.
 
 =cut
+
 use Tie::Hash;
 our @ISA = qw(Tie::ExtraHash Tie::Hash);
 
-use Dpkg::ErrorHandling qw(internerr syserr);
+use Dpkg::ErrorHandling;
 use Dpkg::Gettext;
 
 # Import capit
@@ -66,6 +76,7 @@ Return a reference to a tied hash implementing storage of simple
 "field: value" mapping as used in many Debian-specific files.
 
 =cut
+
 sub new {
     my $hash = {};
     tie %{$hash}, 'Dpkg::Fields::Object';
@@ -138,6 +149,7 @@ following regex: /X[SBC]+-$name/i. Return the name of the field found or
 undef if nothing has been found.
 
 =cut
+
 sub find_custom_field {
     my ($self, $name) = @_;
     foreach my $key (keys %{$self->[0]}) {
@@ -151,6 +163,7 @@ sub find_custom_field {
 Identify a user field and retrieve its value.
 
 =cut
+
 sub get_custom_field {
     my ($self, $name) = @_;
     my $key = $self->find_custom_field($name);
@@ -164,6 +177,7 @@ sub get_custom_field {
 Dump the raw content of the hash either as a string or to a filehandle.
 
 =cut
+
 sub dump {
     my ($self, $fh) = @_;
     my $str = "";
@@ -181,6 +195,7 @@ sub dump {
 Define the order in which fields will be displayed in the output() method.
 
 =cut
+
 sub set_field_importance {
     my ($self, @fields) = @_;
     my $i = 1;
@@ -197,6 +212,7 @@ substitution of variables defined in the Dpkg::Substvars object.
 Also returns the string of what would printed on the filehandle.
 
 =cut
+
 sub output {
     my ($self, $fh, $substvars) = @_;
     my $str = "";
@@ -206,6 +222,7 @@ sub output {
     if (defined($substvars)) {
 	foreach my $f (keys %{$self->[0]}) {
 	    $substvars->set("F:$f", $self->[0]->{$f});
+	    $substvars->no_warn("F:$f");
 	}
     }
 
@@ -228,12 +245,12 @@ sub output {
         }
         $v =~ m/\S/ || next; # delete whitespace-only fields
         $v =~ m/\n\S/ &&
-            internerr(_g("field %s has newline then non whitespace >%s<"),
+            internerr("field %s has newline then non whitespace >%s<",
                       $f, $v);
         $v =~ m/\n[ \t]*\n/ &&
-            internerr(_g("field %s has blank lines >%s<"), $f, $v);
+            internerr("field %s has blank lines >%s<", $f, $v);
         $v =~ m/\n$/ &&
-            internerr(_g("field %s has trailing newline >%s<"), $f, $v);
+            internerr("field %s has trailing newline >%s<", $f, $v);
         if (defined($substvars)) {
            $v =~ s/,[\s,]*,/,/g;
            $v =~ s/^\s*,\s*//;
