@@ -2,8 +2,8 @@
  * dpkg - main program for package management
  * archives.c - actions that process archive files, mainly unpack
  *
- * Copyright (C) 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright (C) 2000 Wichert Akkerman <wakkerma@debian.org>
+ * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2000 Wichert Akkerman <wakkerma@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
@@ -20,6 +20,9 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 #include <config.h>
+#include <compat.h>
+
+#include <dpkg/i18n.h>
 
 #include <errno.h>
 #include <stdio.h>
@@ -38,10 +41,12 @@
 #define obstack_chunk_alloc m_malloc
 #define obstack_chunk_free free
 
-#include <dpkg.h>
-#include <dpkg-db.h>
-#include <tarfn.h>
-#include <myopt.h>
+#include <dpkg/dpkg.h>
+#include <dpkg/dpkg-db.h>
+#include <dpkg/buffer.h>
+#include <dpkg/subproc.h>
+#include <dpkg/tarfn.h>
+#include <dpkg/myopt.h>
 
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
@@ -94,7 +99,7 @@ quote_filename(char *buf, int size, char *s)
       size -= 2;
       break;
     default:
-      if (((*s) & 0x80) == 0) {
+      if (((*s) & 0x80) == '\0') {
         *buf++ = *s++;
         --size;
       } else {
@@ -205,8 +210,7 @@ int filesavespackage(struct fileinlist *file, struct pkginfo *pkgtobesaved,
       debug(dbg_eachfiledetail,"filesavespackage ...  is 3rd package");
 
       if (!thirdpkg->clientdata->fileslistvalid) {
-        debug(dbg_eachfiledetail,
-              _("process_archive ...  already disappeared !"));
+        debug(dbg_eachfiledetail, "process_archive ... already disappeared!");
         continue;
       }
       /* We've found a package that can take this file. */
@@ -230,11 +234,39 @@ int tarfileread(void *ud, char *buf, int len) {
   return r;
 }
 
+static void
+tarfile_skip_one_forward(struct TarInfo *ti,
+                         struct fileinlist **oldnifd,
+                         struct fileinlist *nifd)
+{
+  struct tarcontext *tc = (struct tarcontext *)ti->UserData;
+  size_t r;
+  char databuf[TARBLKSZ];
+
+  obstack_free(&tar_obs, nifd);
+  tc->newfilesp = oldnifd;
+  *oldnifd = NULL;
+
+  /* We need to advance the tar file to the next object, so read the
+   * file data and set it to oblivion.
+   */
+  if ((ti->Type == NormalFile0) || (ti->Type == NormalFile1)) {
+    char fnamebuf[256];
+
+    fd_null_copy(tc->backendpipe, ti->Size,
+                 _("skipped unpacking file '%.255s' (replaced or excluded?)"),
+                 quote_filename(fnamebuf, 256, ti->Name));
+    r = ti->Size % TARBLKSZ;
+    if (r > 0)
+      r = safe_read(tc->backendpipe, databuf, TARBLKSZ - r);
+  }
+}
+
 int fnameidlu;
 struct varbuf fnamevb;
 struct varbuf fnametmpvb;
 struct varbuf fnamenewvb;
-struct packageinlist *deconfigure = NULL;
+struct pkg_deconf_list *deconfigure = NULL;
 
 static time_t currenttime;
 
@@ -373,6 +405,7 @@ int tarobject(struct TarInfo *ti) {
   static struct varbuf conffderefn, hardlinkfn, symlinkfn;
   static int fd;
   const char *usename;
+  struct filenamenode *usenode;
 
   struct conffile *conff;
   struct tarcontext *tc= (struct tarcontext*)ti->UserData;
@@ -423,7 +456,10 @@ int tarobject(struct TarInfo *ti) {
     }
   }
 
-  usename= namenodetouse(nifd->namenode,tc->pkg)->name + 1; /* Skip the leading `/' */
+  usenode = namenodetouse(nifd->namenode, tc->pkg);
+  usename = usenode->name + 1; /* Skip the leading '/'. */
+
+  trig_file_activate(usenode, tc->pkg);
 
   if (nifd->namenode->flags & fnnf_new_conff) {
     /* If it's a conffile we have to extract it next to the installed
@@ -559,9 +595,12 @@ int tarobject(struct TarInfo *ti) {
 	  keepexisting = 1;
         } else {
           if (!statr && S_ISDIR(stab.st_mode)) {
-            forcibleerr(fc_overwritedir, _("trying to overwrite directory `%.250s' "
-                        "in package %.250s with nondirectory"),
-                        nifd->namenode->name,otherpkg->name);
+            forcibleerr(fc_overwritedir,
+                        _("trying to overwrite directory '%.250s' "
+                          "in package %.250s %.250s with nondirectory"),
+                        nifd->namenode->name, otherpkg->name,
+                        versiondescribe(&otherpkg->installed.version,
+                                        vdew_always));
           } else {
             /* WTA: At this point we are replacing something without a Replaces.
 	     * if the new object is a directory and the previous object does not
@@ -569,39 +608,29 @@ int tarobject(struct TarInfo *ti) {
 	     */
             if (! (statr && ti->Type==Directory))
               forcibleerr(fc_overwrite,
-                        _("trying to overwrite `%.250s', which is also in package %.250s"),
-                        nifd->namenode->name,otherpkg->name);
+                          _("trying to overwrite '%.250s', "
+                            "which is also in package %.250s %.250s"),
+                          nifd->namenode->name, otherpkg->name,
+                          versiondescribe(&otherpkg->installed.version,
+                                          vdew_always));
           }
         }
       }
     }
   }
        
+  if (existingdirectory) return 0;
+  if (keepexisting) {
+    tarfile_skip_one_forward(ti, oldnifd, nifd);
+    return 0;
+  }
+
   /* Now, at this stage we want to make sure neither of .dpkg-new and .dpkg-tmp
    * are hanging around.
    */
   ensure_pathname_nonexisting(fnamenewvb.buf);
   ensure_pathname_nonexisting(fnametmpvb.buf);
 
-  if (existingdirectory) return 0;
-  if (keepexisting) {
-    obstack_free(&tar_obs, nifd);
-    tc->newfilesp= oldnifd;
-    *oldnifd = NULL;
-
-    /* We need to advance the tar file to the next object, so read the
-     * file data and set it to oblivion.
-     */
-    if ((ti->Type == NormalFile0) || (ti->Type == NormalFile1)) {
-      char fnamebuf[256];
-      fd_null_copy(tc->backendpipe, ti->Size, _("gobble replaced file `%.255s'"),quote_filename(fnamebuf,256,ti->Name));
-      r= ti->Size % TARBLKSZ;
-      if (r > 0) r= safe_read(tc->backendpipe,databuf,TARBLKSZ - r);
-    }
-
-    return 0;
-  }
-
   /* Now we start to do things that we need to be able to undo
    * if something goes wrong.  Watch out for the CLEANUP comments to
    * keep an eye on what's installed on the disk at each point.
@@ -651,7 +680,8 @@ int tarobject(struct TarInfo *ti) {
      * it until we apply the proper mode, which might be a statoverride.
      */
     fd= open(fnamenewvb.buf, (O_CREAT|O_EXCL|O_WRONLY), 0);
-    if (fd < 0) ohshite(_("unable to create `%.255s'"),ti->Name);
+    if (fd < 0)
+      ohshite(_("unable to create `%.255s' (while processing `%.255s')"), fnamenewvb.buf, ti->Name);
     push_cleanup(cu_closefd, ehflag_bombout, NULL, 0, 1, &fd);
     debug(dbg_eachfiledetail,"tarobject NormalFile[01] open size=%lu",
           (unsigned long)ti->Size);
@@ -729,7 +759,7 @@ int tarobject(struct TarInfo *ti) {
     newtarobject_allmodes(fnamenewvb.buf,ti,nifd->namenode->statoverride);
     break;
   default:
-    internerr("bad tar type, but already checked");
+    internerr("unknown tar type '%d', but already checked", ti->Type);
   }
   /* CLEANUP: Now we have extracted the new object in .dpkg-new (or,
    * if the file already exists as a directory and we were trying to extract
@@ -852,19 +882,16 @@ static int try_deconfigure_can(int (*force_p)(struct deppossi*),
    *                 1: deconfiguration queued ok (no message printed)
    *                 0: not possible (why is printed)
    */
-  struct packageinlist *newdeconf;
+  struct pkg_deconf_list *newdeconf;
   
   if (force_p && force_p(pdep)) {
-    fprintf(stderr, _("dpkg: warning - "
-                      "ignoring dependency problem with %s:\n%s"),
-            action, why);
+    warning(_("ignoring dependency problem with %s:\n%s"), action, why);
     return 2;
   } else if (f_autodeconf) {
     if (pkg->installed.essential) {
       if (fc_removeessential) {
-        fprintf(stderr, _("dpkg: warning - considering deconfiguration of essential\n"
-                          " package %s, to enable %s.\n"),
-                pkg->name, action);
+        warning(_("considering deconfiguration of essential\n"
+                  " package %s, to enable %s."), pkg->name, action);
       } else {
         fprintf(stderr, _("dpkg: no, %s is essential, will not deconfigure\n"
                           " it in order to enable %s.\n"),
@@ -873,10 +900,10 @@ static int try_deconfigure_can(int (*force_p)(struct deppossi*),
       }
     }
     pkg->clientdata->istobe= itb_deconfigure;
-    newdeconf = m_malloc(sizeof(struct packageinlist));
+    newdeconf = m_malloc(sizeof(struct pkg_deconf_list));
     newdeconf->next= deconfigure;
     newdeconf->pkg= pkg;
-    newdeconf->xinfo= removal;
+    newdeconf->pkg_removal = removal;
     deconfigure= newdeconf;
     return 1;
   } else {
@@ -898,11 +925,9 @@ static int try_remove_can(struct deppossi *pdep,
 void check_breaks(struct dependency *dep, struct pkginfo *pkg,
                   const char *pfilename) {
   struct pkginfo *fixbydeconf;
-  struct varbuf why;
+  struct varbuf why = VARBUF_INIT;
   int ok;
 
-  varbufinit(&why);
-
   fixbydeconf = NULL;
   if (depisok(dep, &why, &fixbydeconf, 0)) {
     varbuffree(&why);
@@ -937,8 +962,7 @@ void check_breaks(struct dependency *dep, struct pkginfo *pkg,
   if (ok > 0) return;
 
   if (force_breaks(dep->list)) {
-    fprintf(stderr, _("dpkg: warning - ignoring breakage,"
-                      " may proceed anyway !\n"));
+    warning(_("ignoring breakage, may proceed anyway!"));
     return;
   }
 
@@ -956,11 +980,8 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
                     const char *pfilename) {
   struct pkginfo *fixbyrm;
   struct deppossi *pdep, flagdeppossi;
-  struct varbuf conflictwhy, removalwhy;
+  struct varbuf conflictwhy = VARBUF_INIT, removalwhy = VARBUF_INIT;
   struct dependency *providecheck;
-  
-  varbufinit(&conflictwhy);
-  varbufinit(&removalwhy);
 
   fixbyrm = NULL;
   if (depisok(dep, &conflictwhy, &fixbyrm, 0)) {
@@ -1028,10 +1049,10 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
       if (!pdep && skip_due_to_hold(fixbyrm)) {
         pdep= &flagdeppossi;
       }
-      if (!pdep && (fixbyrm->eflag & eflagf_reinstreq)) {
+      if (!pdep && (fixbyrm->eflag & eflag_reinstreq)) {
         if (fc_removereinstreq) {
           fprintf(stderr, _("dpkg: package %s requires reinstallation, but will"
-                  " remove anyway as you request.\n"), fixbyrm->name);
+                  " remove anyway as you requested.\n"), fixbyrm->name);
         } else {
           fprintf(stderr, _("dpkg: package %s requires reinstallation, "
                   "will not remove.\n"), fixbyrm->name);
@@ -1058,7 +1079,7 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
           pfilename, pkg->name, conflictwhy.buf);
   if (!force_conflicts(dep->list))
     ohshit(_("conflicting packages - not installing %.250s"),pkg->name);
-  fprintf(stderr, _("dpkg: warning - ignoring conflict, may proceed anyway !\n"));
+  warning(_("ignoring conflict, may proceed anyway!"));
   varbuffree(&conflictwhy);
   
   return;
@@ -1067,7 +1088,7 @@ void check_conflict(struct dependency *dep, struct pkginfo *pkg,
 void cu_cidir(int argc, void **argv) {
   char *cidir= (char*)argv[0];
   char *cidirrest= (char*)argv[1];
-  cidirrest[-1]= 0;
+  cidirrest[-1] = '\0';
   ensure_pathname_nonexisting(cidir);
 }  
 
@@ -1158,7 +1179,7 @@ void archivefiles(const char *const *argv) {
     p= findoutput.buf; i=0;
     while (*p) {
       arglist[i++]= p;
-      while ((c= *p++) != 0);
+      while ((c = *p++) != '\0') ;
     }
     arglist[i] = NULL;
     argp= arglist;
@@ -1173,9 +1194,11 @@ void archivefiles(const char *const *argv) {
 
   currenttime = time(NULL);
 
-  varbufinit(&fnamevb);
-  varbufinit(&fnametmpvb);
-  varbufinit(&fnamenewvb);
+  /* Initialize fname variables contents. */
+
+  varbufreset(&fnamevb);
+  varbufreset(&fnametmpvb);
+  varbufreset(&fnamenewvb);
 
   varbufaddstr(&fnamevb,instdir); varbufaddc(&fnamevb,'/');
   varbufaddstr(&fnametmpvb,instdir); varbufaddc(&fnametmpvb,'/');
@@ -1195,8 +1218,8 @@ void archivefiles(const char *const *argv) {
     push_error_handler(&ejbuf,print_error_perpackage,thisarg);
     process_archive(thisarg);
     onerr_abort++;
-    if (ferror(stdout)) werr("stdout");
-    if (ferror(stderr)) werr("stderr");
+    m_output(stdout, _("<standard output>"));
+    m_output(stderr, _("<standard error>"));
     onerr_abort--;
     set_error_display(NULL, NULL);
     error_unwind(ehflag_normaltidy);
@@ -1213,7 +1236,7 @@ void archivefiles(const char *const *argv) {
   case act_avail:
     break;
   default:
-    internerr("unknown action");
+    internerr("unknown action '%d'", cipaction->arg);
   }
 
   trigproc_run_deferred();
@@ -1269,8 +1292,8 @@ int wanttoinstall(struct pkginfo *pkg, const struct versionrevision *ver, int sa
     }
   } else {
     if (fc_downgrade) {
-      if (saywhy) fprintf(stderr, _("%s - warning: downgrading %.250s "
-             "from %.250s to %.250s.\n"), DPKG, pkg->name,
+      if (saywhy)
+        warning(_("downgrading %.250s from %.250s to %.250s."), pkg->name,
              versiondescribe(&pkg->installed.version, vdew_nonambig),
              versiondescribe(&pkg->available.version, vdew_nonambig));
       return 1;
