@@ -2,8 +2,8 @@
  * dpkg - main program for package management
  * configure.c - configure packages
  *
- * Copyright 1995 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright 1999,2002 Wichert Akkerman <wichert@deephackmode.org>
+ * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1999,2002 Wichert Akkerman <wichert@deephackmode.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
@@ -20,6 +20,9 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 #include <config.h>
+#include <compat.h>
+
+#include <dpkg/i18n.h>
 
 #include <errno.h>
 #include <signal.h>
@@ -39,13 +42,19 @@
 #include <time.h>
 #include <sys/termios.h>
 
-#include <dpkg.h>
-#include <dpkg-db.h>
+#include <dpkg/macros.h>
+#include <dpkg/dpkg.h>
+#include <dpkg/dpkg-db.h>
+#include <dpkg/buffer.h>
 
 #include "filesdb.h"
 #include "main.h"
 
-int conffoptcells[2][2]= { CONFFOPTCELLS };
+static int conffoptcells[2][2] = {
+	/* Distro !edited. */	/* Distro edited. */
+	{ cfo_keep,		cfo_install },		/* User !edited. */
+	{ cfo_keep,		cfo_prompt_keep },	/* User edited. */
+};
 
 static void md5hash(struct pkginfo *pkg, char **hashbuf, const char *fn);
 static void copyfileperm(const char* source, const char* target);
@@ -77,7 +86,8 @@ void deferred_configure(struct pkginfo *pkg) {
 	 * Try 4 (only if --force-depends).
 	 *  Do anyway.
 	 */
-	struct varbuf aemsgs, cdr, cdr2;
+	struct varbuf aemsgs = VARBUF_INIT;
+	struct varbuf cdr = VARBUF_INIT, cdr2 = VARBUF_INIT;
 	char *cdr2rest;
 	int ok, r, useredited, distedited;
 	struct conffile *conff;
@@ -99,7 +109,6 @@ void deferred_configure(struct pkginfo *pkg) {
 		if (findbreakcycle(pkg))
 			sincenothing= 0; 
 
-	varbufinit(&aemsgs);
 	ok = dependencies_ok(pkg, NULL, &aemsgs);
 	if (ok == 1) {
 		varbuffree(&aemsgs);
@@ -128,13 +137,13 @@ void deferred_configure(struct pkginfo *pkg) {
 	} else if (aemsgs.used) {
 		varbufaddc(&aemsgs,0);
 		fprintf(stderr,
-				_("dpkg: %s: dependency problems, but configuring anyway as you request:\n%s"),
+				_("dpkg: %s: dependency problems, but configuring anyway as you requested:\n%s"),
 				pkg->name, aemsgs.buf);
 	}
 	varbuffree(&aemsgs);
 	sincenothing= 0;
 
-	if (pkg->eflag & eflagf_reinstreq)
+	if (pkg->eflag & eflag_reinstreq)
 		forcibleerr(fc_removereinstreq,
 				_("Package is in a very bad inconsistent state - you should\n"
 					" reinstall it before attempting configuration."));
@@ -170,10 +179,13 @@ void deferred_configure(struct pkginfo *pkg) {
 		 * If `*.dpkg-new' no longer exists we assume that we've already
 		 * processed this one.
 		 */
-		varbufinit(&cdr);
-		varbufinit(&cdr2);
 		for (conff= pkg->installed.conffiles; conff; conff= conff->next) {
-			r= conffderef(pkg, &cdr, conff->name);
+			struct filenamenode *usenode;
+
+			usenode = namenodetouse(findnamenode(conff->name,
+			                                     fnn_nocopy), pkg);
+
+			r = conffderef(pkg, &cdr, usenode->name);
 			if (r == -1) {
 				conff->hash= EMPTY_HASH;
 				continue;
@@ -209,7 +221,8 @@ void deferred_configure(struct pkginfo *pkg) {
 				what= cfo_identical;
 			} else if (!strcmp(currenthash,NONEXISTENTFLAG) && fc_conff_miss) {
 				fprintf(stderr, _("\nConfiguration file `%s', does not exist on system.\n"
-							"Installing new config file as you request.\n"), conff->name);
+				                  "Installing new config file as you requested.\n"),
+				        usenode->name);
 				what= cfo_newconff;
 				useredited= -1;
 				distedited= -1;
@@ -233,64 +246,61 @@ void deferred_configure(struct pkginfo *pkg) {
 
 			debug(dbg_conff,
 					"deferred_configure `%s' (= `%s') useredited=%d distedited=%d what=%o",
-					conff->name, cdr.buf, useredited, distedited, what);
+			      usenode->name, cdr.buf, useredited, distedited,
+			      what);
 
-			what=promptconfaction(conff->name, cdr.buf, cdr2.buf, useredited, distedited, what);
+			what = promptconfaction(usenode->name, cdr.buf, cdr2.buf,
+			                        useredited, distedited, what);
 
 			switch (what & ~(cfof_isnew|cfof_userrmd)) {
 				case cfo_keep | cfof_backup:
 					strcpy(cdr2rest,DPKGOLDEXT);
 					if (unlink(cdr2.buf) && errno != ENOENT)
-						fprintf(stderr,
-								_("dpkg: %s: warning - failed to remove old backup `%.250s': %s\n"),
-								pkg->name, cdr2.buf, strerror(errno));
+						warning(_("%s: failed to remove old backup '%.250s': %s"),
+						        pkg->name, cdr2.buf, strerror(errno));
 					cdr.used--;
 					varbufaddstr(&cdr,DPKGDISTEXT);
 					varbufaddc(&cdr,0);
 					strcpy(cdr2rest,DPKGNEWEXT);
-					trig_file_activate_byname(conff->name, pkg);
+					trig_file_activate(usenode, pkg);
 					if (rename(cdr2.buf,cdr.buf))
-						fprintf(stderr,
-								_("dpkg: %s: warning - failed to rename `%.250s' to `%.250s': %s\n"),
-								pkg->name, cdr2.buf, cdr.buf, strerror(errno));
+						warning(_("%s: failed to rename '%.250s' to '%.250s': %s"),
+						        pkg->name, cdr2.buf, cdr.buf, strerror(errno));
 					break;
 
 				case cfo_keep:
 					strcpy(cdr2rest,DPKGNEWEXT);
 					if (unlink(cdr2.buf))
-						fprintf(stderr,
-								_("dpkg: %s: warning - failed to remove `%.250s': %s\n"),
-								pkg->name, cdr2.buf, strerror(errno));
+						warning(_("%s: failed to remove '%.250s': %s"),
+						        pkg->name, cdr2.buf, strerror(errno));
 					break;
 
 				case cfo_install | cfof_backup:
 					strcpy(cdr2rest,DPKGDISTEXT);
 					if (unlink(cdr2.buf) && errno != ENOENT)
-						fprintf(stderr,
-								_("dpkg: %s: warning - failed to remove old distrib version `%.250s': %s\n"),
-								pkg->name, cdr2.buf, strerror(errno));
+						warning(_("%s: failed to remove old distrib version '%.250s': %s"),
+						        pkg->name, cdr2.buf, strerror(errno));
 					strcpy(cdr2rest,DPKGOLDEXT);
 					if (unlink(cdr2.buf) && errno != ENOENT)
-						fprintf(stderr,
-								_("dpkg: %s: warning - failed to remove `%.250s' (before overwrite): %s\n"),
-								pkg->name, cdr2.buf, strerror(errno));
+						warning(_("%s: failed to remove '%.250s' (before overwrite): %s"),
+						        pkg->name, cdr2.buf, strerror(errno));
 					if (!(what & cfof_userrmd))
 						if (link(cdr.buf,cdr2.buf))
-							fprintf(stderr,
-								_("dpkg: %s: warning - failed to link `%.250s' to `%.250s': %s\n"),
-								pkg->name, cdr.buf, cdr2.buf, strerror(errno));
+							warning(_("%s: failed to link '%.250s' to '%.250s': %s"),
+							        pkg->name, cdr.buf, cdr2.buf, strerror(errno));
 					/* fall through */
 				case cfo_install:
-					printf(_("Installing new version of config file %s ...\n"),conff->name);
+					printf(_("Installing new version of config file %s ...\n"),
+					       usenode->name);
 				case cfo_newconff:
 					strcpy(cdr2rest,DPKGNEWEXT);
-					trig_file_activate_byname(conff->name, pkg);
+					trig_file_activate(usenode, pkg);
 					if (rename(cdr2.buf,cdr.buf))
 						ohshite(_("unable to install `%.250s' as `%.250s'"),cdr2.buf,cdr.buf);
 					break;
 
 				default:
-					internerr("unknown what");
+					internerr("unknown conffopt '%d'", what);
 			}
 
 			conff->hash= nfstrsave(newdisthash);
@@ -316,7 +326,7 @@ void deferred_configure(struct pkginfo *pkg) {
 	                           NULL))
 		putchar('\n');
 
-	pkg->eflag= eflagv_ok;
+	pkg->eflag = eflag_ok;
 	post_postinst_tasks(pkg, stat_installed);
 }
 
@@ -344,9 +354,9 @@ int conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in) {
 		debug(dbg_conffdetail,"conffderef in=`%s' current working=`%s'", in, result->buf);
 		if (lstat(result->buf,&stab)) {
 			if (errno != ENOENT)
-				fprintf(stderr, _("dpkg: %s: warning - unable to stat config file `%s'\n"
-							" (= `%s'): %s\n"),
-						pkg->name, in, result->buf, strerror(errno));
+				warning(_("%s: unable to stat config file '%s'\n"
+				          " (= '%s'): %s"),
+				        pkg->name, in, result->buf, strerror(errno));
 			debug(dbg_conffdetail,"conffderef nonexistent");
 			return 0;
 		} else if (S_ISREG(stab.st_mode)) {
@@ -355,8 +365,8 @@ int conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in) {
 		} else if (S_ISLNK(stab.st_mode)) {
 			debug(dbg_conffdetail,"conffderef symlink loopprotect=%d",loopprotect);
 			if (loopprotect++ >= 25) {
-				fprintf(stderr, _("dpkg: %s: warning - config file `%s' is a circular link\n"
-							" (= `%s')\n"), pkg->name, in, result->buf);
+				warning(_("%s: config file '%s' is a circular link\n"
+				          " (= '%s')"), pkg->name, in, result->buf);
 				return -1;
 			}
 			need= 255;
@@ -368,17 +378,17 @@ int conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in) {
 				}
 				r= readlink(result->buf,linkreadbuf,linkreadbufsize-1);
 				if (r < 0) {
-					fprintf(stderr, _("dpkg: %s: warning - unable to readlink conffile `%s'\n"
-								" (= `%s'): %s\n"),
-							pkg->name, in, result->buf, strerror(errno));
+					warning(_("%s: unable to readlink conffile '%s'\n"
+					          " (= '%s'): %s"),
+					        pkg->name, in, result->buf, strerror(errno));
 					return -1;
 				}
 				debug(dbg_conffdetail,"conffderef readlink gave %d, `%.*s'",
-						r, r>0 ? r : 0, linkreadbuf);
+						r, max(r, 0), linkreadbuf);
 				if (r < linkreadbufsize-1) break;
 				need= r<<2;
 			}
-			linkreadbuf[r]= 0;
+			linkreadbuf[r] = '\0';
 			if (linkreadbuf[0] == '/') {
 				varbufreset(result);
 				varbufaddstr(result,instdir);
@@ -387,10 +397,9 @@ int conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in) {
 				for (r=result->used-2; r>0 && result->buf[r] != '/'; r--)
 					;
 				if (r < 0) {
-					fprintf(stderr,
-							_("dpkg: %s: warning - conffile `%.250s' resolves to degenerate filename\n"
-								" (`%s' is a symlink to `%s')\n"),
-							pkg->name, in, result->buf, linkreadbuf);
+					warning(_("%s: conffile '%.250s' resolves to degenerate filename\n"
+					          " ('%s' is a symlink to '%s')"),
+					        pkg->name, in, result->buf, linkreadbuf);
 					return -1;
 				}
 				if (result->buf[r] == '/') r++;
@@ -401,9 +410,8 @@ int conffderef(struct pkginfo *pkg, struct varbuf *result, const char *in) {
 			varbufaddstr(result,linkreadbuf);
 			varbufaddc(result,0);
 		} else {
-			fprintf(stderr, _("dpkg: %s: warning - conffile `%.250s' is not a plain"
-						" file or symlink (= `%s')\n"),
-					pkg->name, in, result->buf);
+			warning(_("%s: conffile '%.250s' is not a plain file or symlink (= '%s')"),
+			        pkg->name, in, result->buf);
 			return -1;
 		}
 	}
@@ -425,8 +433,8 @@ static void md5hash(struct pkginfo *pkg, char **hashbuf, const char *fn) {
 	} else if (errno==ENOENT) {
 		*hashbuf = m_strdup(NONEXISTENTFLAG);
 	} else {
-		fprintf(stderr, _("dpkg: %s: warning - unable to open conffile %s for hash: %s\n"),
-				pkg->name, fn, strerror(errno));
+		warning(_("%s: unable to open conffile %s for hash: %s"),
+		        pkg->name, fn, strerror(errno));
 		*hashbuf = m_strdup("-");
 	}
 }
@@ -639,7 +647,7 @@ static enum conffopt promptconfaction(const char* cfgfile, const char* realold,
 			else if (what & cfof_install) { cc= 'y'; break; }
 		}
 
-		/* fixme: say something if silently not install */
+		/* FIXME: say something if silently not install */
 		if (cc == 'd')
 			showdiff(realold, realnew);
 
@@ -665,7 +673,7 @@ static enum conffopt promptconfaction(const char* cfgfile, const char* realold,
 			break;
 
 		default:
-			internerr("unknown response");
+			internerr("unknown response '%d'", cc);
 	}
 
 	return what;
