@@ -2,7 +2,7 @@
  * dpkg - main program for package management
  * depcon.c - dependency and conflict checking
  *
- * Copyright (C) 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
@@ -19,6 +19,9 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 #include <config.h>
+#include <compat.h>
+
+#include <dpkg/i18n.h>
 
 #include <errno.h>
 #include <unistd.h>
@@ -26,8 +29,8 @@
 #include <sys/types.h>
 #include <assert.h>
 
-#include <dpkg.h>
-#include <dpkg-db.h>
+#include <dpkg/dpkg.h>
+#include <dpkg/dpkg-db.h>
 
 #include "main.h"
 
@@ -56,7 +59,8 @@ static int foundcyclebroken(struct cyclesofarlink *thislink,
   for (sol=sofar; sol && sol->pkg != dependedon; sol=sol->back);
 
   /* If not, we do a recursive search on it to see what we find. */
-  if (!sol) return findbreakcyclerecursive(dependedon,thislink);
+  if (!sol)
+    return findbreakcyclerecursive(dependedon, thislink);
   
   debug(dbg_depcon,"found cycle");
   /* Right, we now break one of the links.  We prefer to break
@@ -101,13 +105,20 @@ static int findbreakcyclerecursive(struct pkginfo *pkg, struct cyclesofarlink *s
   pkg->color = gray;
   
   if (f_debug & dbg_depcondetail) {
-    fprintf(stderr,"D0%05o: findbreakcyclerecursive %s ",dbg_depcondetail,pkg->name);
-    for (sol=sofar; sol; sol=sol->back) fprintf(stderr," <- %s",sol->pkg->name);
-    fprintf(stderr,"\n");
+    struct varbuf str_pkgs = VARBUF_INIT;
+
+    for (sol = sofar; sol; sol = sol->back) {
+      varbufaddstr(&str_pkgs, " <- ");
+      varbufaddstr(&str_pkgs, sol->pkg->name);
+    }
+    varbufaddc(&str_pkgs, '\0');
+    debug(dbg_depcondetail, "findbreakcyclerecursive %s %s", pkg->name,
+          str_pkgs.buf);
+    varbuffree(&str_pkgs);
   }
   thislink.pkg= pkg;
   thislink.back= sofar;
-  thislink.possi= 0;
+  thislink.possi = NULL;
   for (dep= pkg->installed.depends; dep; dep= dep->next) {
     if (dep->type != dep_depends && dep->type != dep_predepends) continue;
     for (possi= dep->list; possi; possi= possi->next) {
@@ -142,22 +153,46 @@ int findbreakcycle(struct pkginfo *pkg) {
   for (iter = iterpkgstart(); (tpkg=iterpkgnext(iter)); ) {
     tpkg->color = white;
   }
+  iterpkgend(iter);
 
-  return findbreakcyclerecursive(pkg, 0);
+  return findbreakcyclerecursive(pkg, NULL);
 }
 
 void describedepcon(struct varbuf *addto, struct dependency *dep) {
-  varbufaddstr(addto,dep->up->name);
+  const char *fmt;
+  struct varbuf depstr = VARBUF_INIT;
+
   switch (dep->type) {
-  case dep_depends:     varbufaddstr(addto, _(" depends on "));     break;
-  case dep_predepends:  varbufaddstr(addto, _(" pre-depends on ")); break;
-  case dep_recommends:  varbufaddstr(addto, _(" recommends "));     break;
-  case dep_conflicts:   varbufaddstr(addto, _(" conflicts with ")); break;
-  case dep_suggests:    varbufaddstr(addto, _(" suggests ")); break;
-  case dep_enhances:    varbufaddstr(addto, _(" enhances "));       break;
-  default:              internerr("unknown deptype");
+  case dep_depends:
+    fmt = _("%s depends on %s");
+    break;
+  case dep_predepends:
+    fmt = _("%s pre-depends on %s");
+    break;
+  case dep_recommends:
+    fmt = _("%s recommends %s");
+    break;
+  case dep_suggests:
+    fmt = _("%s suggests %s");
+    break;
+  case dep_breaks:
+    fmt = _("%s breaks %s");
+    break;
+  case dep_conflicts:
+    fmt = _("%s conflicts with %s");
+    break;
+  case dep_enhances:
+    fmt = _("%s enhances %s");
+    break;
+  default:
+    internerr("unknown deptype '%d'", dep->type);
   }
-  varbufdependency(addto, dep);
+
+  varbufdependency(&depstr, dep);
+  varbufaddc(&depstr, 0);
+
+  varbufprintf(addto, fmt, dep->up->name, depstr.buf);
+  varbuffree(&depstr);
 }
   
 int depisok(struct dependency *dep, struct varbuf *whynot,
@@ -171,6 +206,10 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
    * before a package is unpacked, when it is sufficient for the package
    * to be unpacked provided that both the unpacked and previously-configured
    * versions are acceptable.
+   * On 0 return (`not OK'), *canfixbyremove refers to a package which
+   * if removed (dep_conflicts) or deconfigured (dep_breaks) will fix
+   * the problem.  Caller may pass 0 for canfixbyremove and need not
+   * initialise *canfixbyremove.
    */
   struct deppossi *possi;
   struct deppossi *provider;
@@ -184,9 +223,13 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
   char linebuf[1024];
 
   assert(dep->type == dep_depends || dep->type == dep_predepends ||
-         dep->type == dep_conflicts || dep->type == dep_recommends ||
-	 dep->type == dep_suggests || dep->type == dep_enhances );
+	 dep->type == dep_breaks || dep->type == dep_conflicts ||
+	 dep->type == dep_recommends || dep->type == dep_suggests ||
+	 dep->type == dep_enhances);
   
+  if (canfixbyremove)
+    *canfixbyremove = NULL;
+
   /* The dependency is always OK if we're trying to remove the depend*ing*
    * package.
    */
@@ -194,21 +237,23 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
   case itb_remove: case itb_deconfigure:
     return 1;
   case itb_normal:
-    /* Only `installed' packages can be make dependency problems */
+    /* Only installed packages can be make dependency problems */
     switch (dep->up->status) {
     case stat_installed:
+    case stat_triggerspending:
+    case stat_triggersawaited:
       break;
     case stat_notinstalled: case stat_configfiles: case stat_halfinstalled:
     case stat_halfconfigured: case stat_unpacked:
       return 1;
     default:
-      internerr("unknown status depending");
+      internerr("unknown status depending '%d'", dep->up->status);
     }
     break;
   case itb_installnew: case itb_preinstall:
     break;
   default:
-    internerr("unknown istobe depending");
+    internerr("unknown istobe depending '%d'", dep->up->clientdata->istobe);
   }
 
   /* Describe the dependency, in case we have to moan about it. */
@@ -244,6 +289,7 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
       case itb_normal: case itb_preinstall:
         switch (possi->ed->status) {
         case stat_installed:
+        case stat_triggerspending:
           if (versionsatisfied(&possi->ed->installed,possi)) return 1;
           sprintf(linebuf,_("  %.250s is installed, but is version %.250s.\n"),
                   possi->ed->name,
@@ -254,10 +300,11 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
            * Later on, if nothing has put anything in linebuf, we know that it
            * isn't and issue a diagnostic then.
            */
-          *linebuf= 0;
+          *linebuf = '\0';
           break;
         case stat_unpacked:
         case stat_halfconfigured:
+        case stat_triggersawaited:
           if (allowunconfigd) {
             if (!informativeversion(&possi->ed->configversion)) {
               sprintf(linebuf, _("  %.250s is unpacked, but has never been configured.\n"),
@@ -278,6 +325,7 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
               return 1;
             }
           }
+          /* Fall through. */
         default:
           sprintf(linebuf, _("  %.250s is %s.\n"),
                   possi->ed->name, gettext(statusstrings[possi->ed->status]));
@@ -285,7 +333,7 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
         }
         break;
       default:
-        internerr("unknown istobe depended");
+        internerr("unknown istobe depended '%d'", possi->ed->clientdata->istobe);
       }
       varbufaddstr(whynot, linebuf);
 
@@ -329,7 +377,8 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
                     gettext(statusstrings[provider->up->up->status]));
             break;
           default:
-            internerr("unknown istobe provider");
+            internerr("unknown istobe provider '%d'",
+                      provider->up->up->clientdata->istobe);
           }
           varbufaddstr(whynot, linebuf);
         }
@@ -344,12 +393,11 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
       }
     }
 
-    if (canfixbyremove) *canfixbyremove= 0;
     return 0;
 
   } else {
     
-    /* It's a conflict.  There's only one main alternative,
+    /* It's conflicts or breaks.  There's only one main alternative,
      * but we also have to consider Providers.  We return `0' as soon
      * as we find something that matches the conflict, and only describe
      * it then.  If we get to the end without finding anything we return `1'.
@@ -359,9 +407,10 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
     nconflicts= 0;
 
     if (possi->ed != possi->up->up) {
-      /* If the package conflicts with itself it must mean that it conflicts
-       * with other packages which provide the same virtual name.  We therefore
-       * don't look at the real package and go on to the virtual ones.
+      /* If the package conflicts with or breaks itself it must mean
+       * other packages which provide the same virtual name.  We
+       * therefore don't look at the real package and go on to the
+       * virtual ones.
        */
       
       switch (possi->ed->clientdata->istobe) {
@@ -377,13 +426,21 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
         nconflicts++;
         *canfixbyremove= possi->ed;
         break;
-      case itb_normal: case itb_deconfigure: case itb_preinstall:
+      case itb_deconfigure:
+        if (dep->type == dep_breaks) break; /* already deconfiguring this */
+        /* fall through */
+      case itb_normal: case itb_preinstall:
         switch (possi->ed->status) {
         case stat_notinstalled: case stat_configfiles:
           break;
-        default:
+        case stat_halfinstalled: case stat_unpacked:
+        case stat_halfconfigured:
+          if (dep->type == dep_breaks) break; /* no problem */
+        case stat_installed:
+        case stat_triggerspending:
+        case stat_triggersawaited:
           if (!versionsatisfied(&possi->ed->installed, possi)) break;
-          sprintf(linebuf, _("  %.250s (version %.250s) is %s.\n"),
+          sprintf(linebuf, _("  %.250s (version %.250s) is present and %s.\n"),
                   possi->ed->name,
                   versiondescribe(&possi->ed->installed.version,vdew_nonambig),
                   gettext(statusstrings[possi->ed->status]));
@@ -394,7 +451,8 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
         }
         break;
       default:
-        internerr("unknown istobe conflict");
+        internerr("unknown istobe conflict '%d'",
+                  possi->ed->clientdata->istobe);
       }
     }
 
@@ -412,7 +470,8 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
                 provider->up->up->name, possi->ed->name);
         varbufaddstr(whynot, linebuf);
         /* We can't remove the one we're about to install: */
-        if (canfixbyremove) *canfixbyremove= 0;
+        if (canfixbyremove)
+          *canfixbyremove = NULL;
         return 0;
       }
 
@@ -434,12 +493,20 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
           continue; 
         case itb_remove:
           continue;
-        case itb_normal: case itb_deconfigure: case itb_preinstall:
+        case itb_deconfigure:
+          if (dep->type == dep_breaks) continue; /* already deconfiguring */
+        case itb_normal: case itb_preinstall:
           switch (provider->up->up->status) {
           case stat_notinstalled: case stat_configfiles:
             continue;
-          default:
-            sprintf(linebuf, _("  %.250s provides %.250s and is %s.\n"),
+          case stat_halfinstalled: case stat_unpacked:
+          case stat_halfconfigured:
+            if (dep->type == dep_breaks) break; /* no problem */
+          case stat_installed:
+          case stat_triggerspending:
+          case stat_triggersawaited:
+            sprintf(linebuf,
+                    _("  %.250s provides %.250s and is present and %s.\n"),
                     provider->up->up->name, possi->ed->name,
                     gettext(statusstrings[provider->up->up->status]));
             varbufaddstr(whynot, linebuf);
@@ -450,13 +517,15 @@ int depisok(struct dependency *dep, struct varbuf *whynot,
           }
           break;
         default:
-          internerr("unknown istobe conflict provider");
+          internerr("unknown istobe conflict provider '%d'",
+                    provider->up->up->clientdata->istobe);
         }
       }
     }
 
     if (!nconflicts) return 1;
-    if (nconflicts>1) *canfixbyremove= 0;
+    if (nconflicts > 1)
+      *canfixbyremove = NULL;
     return 0;
 
   } /* if (dependency) {...} else {...} */
