@@ -2,7 +2,7 @@
  * dpkg - main program for package management
  * help.c - various helper routines
  *
- * Copyright (C) 1995 Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 1995 Ian Jackson <ian@chiark.greenend.org.uk>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as
@@ -19,8 +19,12 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 #include <config.h>
+#include <compat.h>
+
+#include <dpkg/i18n.h>
 
 #include <errno.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include <dirent.h>
 #include <assert.h>
@@ -28,11 +32,12 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
-#include <signal.h>
 #include <time.h>
 
-#include <dpkg.h>
-#include <dpkg-db.h>
+#include <dpkg/dpkg.h>
+#include <dpkg/dpkg-db.h>
+#include <dpkg/path.h>
+#include <dpkg/subproc.h>
 
 #include "filesdb.h"
 #include "main.h"
@@ -53,7 +58,7 @@ struct filenamenode *namenodetouse(struct filenamenode *namenode, struct pkginfo
   
   if (!namenode->divert) {
     r = namenode;
-    goto found;
+    return r;
   }
   
   debug(dbg_eachfile,"namenodetouse namenode=`%s' pkg=%s",
@@ -70,8 +75,6 @@ struct filenamenode *namenodetouse(struct filenamenode *namenode, struct pkginfo
         namenode->divert->pkg ? namenode->divert->pkg->name : "<none>",
         r->name);
 
- found:
-  trig_file_activate(r, pkg);
   return r;
 }
 
@@ -122,7 +125,7 @@ void checkpath(void) {
 }
 
 int ignore_depends(struct pkginfo *pkg) {
-  struct packageinlist *id;
+  struct pkginqueue *id;
   for (id= ignoredependss; id; id= id->next)
     if (id->pkg == pkg) return 1;
   return 0;
@@ -146,19 +149,26 @@ int force_conflicts(struct deppossi *possi) {
 
 static const char* preexecscript(const char *path, char *const *argv) {
   /* returns the path to the script inside the chroot
-   * none of the stuff here will work if admindir isn't inside instdir
-   * as expected. - fixme
+   * FIXME: none of the stuff here will work if admindir isn't inside
+   * instdir as expected.
    */
   size_t instdirl;
 
   if (*instdir) {
     if (chroot(instdir)) ohshite(_("failed to chroot to `%.250s'"),instdir);
-    if (chdir("/")) ohshite(_("failed to chdir to `%.255s'"),"/");
+    if (chdir("/"))
+      ohshite(_("failed to chdir to `%.255s'"), "/");
   }
   if (f_debug & dbg_scripts) {
-    fprintf(stderr,"D0%05o: fork/exec %s (",dbg_scripts,path);
-    while (*++argv) fprintf(stderr," %s",*argv);
-    fputs(" )\n",stderr);
+    struct varbuf args = VARBUF_INIT;
+
+    while (*++argv) {
+      varbufaddc(&args, ' ');
+      varbufaddstr(&args, *argv);
+    }
+    varbufaddc(&args, '\0');
+    debug(dbg_scripts, "fork/exec %s (%s )", path, args.buf);
+    varbuffree(&args);
   }
   instdirl= strlen(instdir);
   if (!instdirl) return path;
@@ -194,38 +204,6 @@ static char *const *buildarglist(const char *scriptname, ...) {
   return arglist;
 }
 
-#define NSCRIPTCATCHSIGNALS (int)(sizeof(script_catchsignallist)/sizeof(int)-1)
-static int script_catchsignallist[]= { SIGQUIT, SIGINT, 0 };
-static struct sigaction script_uncatchsignal[NSCRIPTCATCHSIGNALS];
-
-static void cu_restorescriptsignals(int argc, void **argv) {
-  int i;
-  for (i=0; i<NSCRIPTCATCHSIGNALS; i++) {
-    if (sigaction(script_catchsignallist[i], &script_uncatchsignal[i], NULL)) {
-      fprintf(stderr,_("error un-catching signal %s: %s\n"),
-              strsignal(script_catchsignallist[i]),strerror(errno));
-      onerr_abort++;
-    }
-  }
-}
-
-static void script_catchsignals(void) {
-  int i;
-  struct sigaction catchsig;
-  
-  onerr_abort++;
-  memset(&catchsig,0,sizeof(catchsig));
-  catchsig.sa_handler= SIG_IGN;
-  sigemptyset(&catchsig.sa_mask);
-  catchsig.sa_flags= 0;
-  for (i=0; i<NSCRIPTCATCHSIGNALS; i++)
-    if (sigaction(script_catchsignallist[i],&catchsig,&script_uncatchsignal[i]))
-      ohshite(_("unable to ignore signal %s before running script"),
-              strsignal(script_catchsignallist[i]));
-  push_cleanup(cu_restorescriptsignals, ~0, NULL, 0, 0);
-  onerr_abort--;
-}
-
 void
 post_postinst_tasks(struct pkginfo *pkg, enum pkgstatus new_status)
 {
@@ -267,7 +245,12 @@ static void setexecute(const char *path, struct stat *stab) {
   if (!chmod(path,0755)) return;
   ohshite(_("unable to set execute permissions on `%.250s'"),path);
 }
-static int do_script(const char *pkg, const char *scriptname, const char *scriptpath, struct stat *stab, char *const arglist[], const char *desc, const char *name, int warn) {
+
+static int
+do_script(struct pkginfo *pkg, struct pkginfoperfile *pif,
+          const char *scriptname, const char *scriptpath, struct stat *stab,
+          char *const arglist[], const char *desc, const char *name, int warn)
+{
   const char *scriptexec;
   int c1, r;
   setexecute(scriptpath,stab);
@@ -283,13 +266,14 @@ static int do_script(const char *pkg, const char *scriptname, const char *script
       narglist[r]= arglist[r];
     scriptexec= preexecscript(scriptpath,(char * const *)narglist);
     narglist[0]= scriptexec;
-    if (setenv(MAINTSCRIPTPKGENVVAR, pkg, 1) ||
+    if (setenv(MAINTSCRIPTPKGENVVAR, pkg->name, 1) ||
+        setenv(MAINTSCRIPTARCHENVVAR, pif->architecture, 1) ||
         setenv(MAINTSCRIPTDPKGENVVAR, PACKAGE_VERSION, 1))
-      ohshite(_("unable to setenv for maint script"));
+      ohshite(_("unable to setenv for maintainer script"));
     execv(scriptexec,(char * const *)narglist);
     ohshite(desc,name);
   }
-  script_catchsignals(); /* This does a push_cleanup() */
+  setup_subproc_signals(name); /* This does a push_cleanup() */
   r= waitsubproc(c1,name,warn);
   pop_cleanup(ehflag_normaltidy);
 
@@ -309,7 +293,7 @@ vmaintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
 
   scriptpath= pkgadminfile(pkg,scriptname);
   arglist= vbuildarglist(scriptname,ap);
-  sprintf(buf,"%s script",description);
+  sprintf(buf, _("installed %s script"), description);
 
   if (stat(scriptpath,&stab)) {
     if (errno == ENOENT) {
@@ -317,9 +301,10 @@ vmaintainer_script_installed(struct pkginfo *pkg, const char *scriptname,
             scriptname);
       return 0;
     }
-    ohshite(_("unable to stat installed %s script `%.250s'"),description,scriptpath);
+    ohshite(_("unable to stat %s `%.250s'"), buf, scriptpath);
   }
-  do_script(pkg->name, scriptname, scriptpath, &stab, arglist, _("unable to execute %s"), buf, 0);
+  do_script(pkg, &pkg->installed, scriptname, scriptpath, &stab,
+            arglist, _("unable to execute %s"), buf, 0);
 
   return 1;
 }
@@ -356,9 +341,11 @@ maintainer_script_postinst(struct pkginfo *pkg, ...)
   return r;
 }
 
-int maintainer_script_new(const char *pkgname,
-			  const char *scriptname, const char *description,
-                          const char *cidir, char *cidirrest, ...) {
+int
+maintainer_script_new(struct pkginfo *pkg,
+                      const char *scriptname, const char *description,
+                      const char *cidir, char *cidirrest, ...)
+{
   char *const *arglist;
   struct stat stab;
   va_list ap;
@@ -367,7 +354,7 @@ int maintainer_script_new(const char *pkgname,
   va_start(ap,cidirrest);
   arglist= vbuildarglist(scriptname,ap);
   va_end(ap);
-  sprintf(buf,"%s script",description);
+  sprintf(buf, _("new %s script"), description);
 
   strcpy(cidirrest,scriptname);
   if (stat(cidir,&stab)) {
@@ -375,9 +362,10 @@ int maintainer_script_new(const char *pkgname,
       debug(dbg_scripts,"maintainer_script_new nonexistent %s `%s'",scriptname,cidir);
       return 0;
     }
-    ohshite(_("unable to stat new %s script `%.250s'"),description,cidir);
+    ohshite(_("unable to stat %s `%.250s'"), buf, cidir);
   }
-  do_script(pkgname, scriptname, cidir, &stab, arglist, _("unable to execute new %s"), buf, 0);
+  do_script(pkg, &pkg->available, scriptname, cidir, &stab,
+            arglist, _("unable to execute %s"), buf, 0);
   post_script_tasks();
 
   return 1;
@@ -404,12 +392,11 @@ int maintainer_script_alternative(struct pkginfo *pkg,
             scriptname,oldscriptpath);
       return 0;
     }
-    fprintf(stderr,
-            _("dpkg: warning - unable to stat %s `%.250s': %s\n"),
+    warning(_("unable to stat %s '%.250s': %s"),
             buf,oldscriptpath,strerror(errno));
   } else {
-    if (!do_script(pkg->name, scriptname, oldscriptpath, &stab, arglist,
-                   _("unable to execute %s"), buf, PROCWARN)) {
+    if (!do_script(pkg, &pkg->installed, scriptname, oldscriptpath, &stab,
+                   arglist, _("unable to execute %s"), buf, PROCWARN)) {
       post_script_tasks();
       return 1;
     }
@@ -430,7 +417,8 @@ int maintainer_script_alternative(struct pkginfo *pkg,
       ohshite(_("unable to stat %s `%.250s'"),buf,cidir);
   }
 
-  do_script(pkg->name, scriptname, cidir, &stab, arglist, _("unable to execute %s"), buf, 0);
+  do_script(pkg, &pkg->available, scriptname, cidir, &stab,
+            arglist, _("unable to execute %s"), buf, 0);
   fprintf(stderr, _("dpkg: ... it looks like that went OK.\n"));
 
   post_script_tasks();
@@ -546,7 +534,7 @@ void ensure_pathname_nonexisting(const char *pathname) {
   int c1;
   const char *u, *failed;
 
-  u= skip_slash_dotslash(pathname);
+  u = path_skip_slash_dotslash(pathname);
   assert(*u);
 
   debug(dbg_eachfile,"ensure_pathname_nonexisting `%s'",pathname);
@@ -561,9 +549,9 @@ void ensure_pathname_nonexisting(const char *pathname) {
     if (errno == ENOTDIR) return;
   }
   if (errno != ENOTEMPTY && errno != EEXIST) { /* Huh ? */
-    char mbuf[250];
-    snprintf(mbuf, sizeof(mbuf), N_("failed to %s `%%.255s'"), failed);
-    ohshite(_(mbuf),pathname);
+    const char *failed_local = gettext(failed);
+
+    ohshite(_("failed to %s '%.255s'"), failed_local, pathname);
   }
   c1= m_fork();
   if (!c1) {
